@@ -1,12 +1,15 @@
+use std::io::Write;
+
 use binread::{
     io::{Read, Seek, SeekFrom},
     BinRead, BinResult, ReadOptions,
 };
 
 #[cfg(feature = "serde")]
-use serde::{Deserialize, Serialize, Serializer};
+use serde::{Deserialize, Serialize};
+use ssbh_write::SsbhWrite;
 
-use crate::absolute_offset_checked;
+use crate::{absolute_offset_checked, round_up, write_relative_offset};
 
 // Array element types vary in size, so pick a more consersative value.
 const SSBH_ARRAY_MAX_INITIAL_CAPACITY: usize = 1024;
@@ -69,7 +72,7 @@ where
 #[cfg(feature = "hex_buffer")]
 fn serialize_hex<S>(bytes: &[u8], serializer: S) -> Result<S::Ok, S::Error>
 where
-    S: Serializer,
+    S: serde::Serializer,
 {
     serializer.serialize_str(&hex::encode(bytes))
 }
@@ -162,6 +165,7 @@ fn read_elements<C: Copy + 'static, BR: BinRead<Args = C>, R: Read + Seek>(
     // Reduce the risk of failed allocations due to malformed array lengths (ex: -1 in two's complement).
     // This only bounds the initial capacity, so large elements can still resize the vector as needed.
     // This won't impact performance or memory usage for array lengths within the bound.
+    // TODO: Use try_reserve from rust 1.57+ for added stability on windows?
     let mut elements = Vec::with_capacity(std::cmp::min(
         count as usize,
         SSBH_ARRAY_MAX_INITIAL_CAPACITY,
@@ -227,6 +231,89 @@ fn read_ssbh_array<
     result
 }
 
+fn write_array_header<W: Write + Seek>(
+    writer: &mut W,
+    data_ptr: &mut u64,
+    count: usize,
+) -> std::io::Result<()> {
+    // Arrays are always 8 byte aligned.
+    *data_ptr = round_up(*data_ptr, 8);
+
+    // Don't write the offset for empty arrays.
+    if count == 0 {
+        u64::write(&0u64, writer)?;
+    } else {
+        write_relative_offset(writer, data_ptr)?;
+    }
+
+    (count as u64).write(writer)?;
+    Ok(())
+}
+
+impl SsbhWrite for SsbhByteBuffer {
+    fn ssbh_write<W: Write + Seek>(
+        &self,
+        writer: &mut W,
+        data_ptr: &mut u64,
+    ) -> std::io::Result<()> {
+        let current_pos = writer.stream_position()?;
+        if *data_ptr < current_pos + self.size_in_bytes() {
+            *data_ptr = current_pos + self.size_in_bytes();
+        }
+
+        write_array_header(writer, data_ptr, self.elements.len())?;
+
+        let current_pos = writer.stream_position()?;
+        writer.seek(SeekFrom::Start(*data_ptr))?;
+        // Use a custom implementation to avoid writing bytes individually.
+        // Pointers in array elements should point past the end of the array.
+        writer.write_all(&self.elements)?;
+        *data_ptr += self.elements.len() as u64;
+
+        writer.seek(SeekFrom::Start(current_pos))?;
+        Ok(())
+    }
+
+    fn size_in_bytes(&self) -> u64 {
+        16
+    }
+}
+
+impl<T: binread::BinRead + SsbhWrite + Sized> SsbhWrite for SsbhArray<T> {
+    fn ssbh_write<W: Write + Seek>(
+        &self,
+        writer: &mut W,
+        data_ptr: &mut u64,
+    ) -> std::io::Result<()> {
+        // TODO: Create a macro or function for this?
+        let current_pos = writer.stream_position()?;
+        if *data_ptr < current_pos + self.size_in_bytes() {
+            *data_ptr = current_pos + self.size_in_bytes();
+        }
+
+        write_array_header(writer, data_ptr, self.elements.len())?;
+
+        let pos_after_length = writer.stream_position()?;
+        writer.seek(SeekFrom::Start(*data_ptr))?;
+
+        self.elements.as_slice().ssbh_write(writer, data_ptr)?;
+
+        writer.seek(SeekFrom::Start(pos_after_length))?;
+
+        Ok(())
+    }
+
+    fn size_in_bytes(&self) -> u64 {
+        // A 64 bit relative offset and 64 bit length
+        16
+    }
+
+    fn alignment_in_bytes() -> u64 {
+        // Arrays are always 8 byte aligned.
+        8
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use binread::BinReaderExt;
@@ -234,6 +321,8 @@ mod tests {
 
     use hexlit::hex;
 
+    use crate::SsbhString;
+
     use super::*;
 
     #[test]
@@ -299,6 +388,21 @@ mod tests {
         assert_eq!(1u16, value);
     }
 
+    #[test]
+    #[ignore]
+    fn read_ssbh_array_null_nonzero_count() {
+        // TODO: How would in game parsers handle this case?
+        let mut reader = Cursor::new(hex!(
+            "00000000 00000000 03000000 00000000 01000200 03000400"
+        ));
+        let value = reader.read_le::<SsbhArray<u16>>().unwrap();
+        assert_eq!(Vec::<u16>::new(), value.elements);
+
+        // Make sure the reader position is restored.
+        let value = reader.read_le::<u16>().unwrap();
+        assert_eq!(1u16, value);
+    }
+
     #[test]
     fn read_ssbh_array_offset_overflow() {
         let mut reader = Cursor::new(hex!(
@@ -418,4 +522,88 @@ mod tests {
         let value = reader.read_le::<u16>().unwrap();
         assert_eq!(1u16, value);
     }
+
+    #[test]
+    fn ssbh_write_array_ssbh_string() {
+        let value = SsbhArray::new(vec![
+            SsbhString::from("leyes_eye_mario_l_col"),
+            SsbhString::from("eye_mario_w_nor"),
+        ]);
+
+        let mut writer = Cursor::new(Vec::new());
+        let mut data_ptr = 0;
+        value.ssbh_write(&mut writer, &mut data_ptr).unwrap();
+
+        // Check that the relative offsets point past the array.
+        // Check that string data is aligned to 4.
+        assert_eq!(
+            writer.into_inner(),
+            hex!(
+                "10000000 00000000 02000000 00000000
+                 10000000 00000000 20000000 00000000
+                 6C657965 735F6579 655F6D61 72696F5F 
+                 6C5F636F 6C000000 6579655F 6D617269 
+                 6F5F775F 6E6F7200"
+            )
+        );
+    }
+
+    #[test]
+    fn write_empty_array() {
+        let value = SsbhArray::<u32>::new(Vec::new());
+
+        let mut writer = Cursor::new(Vec::new());
+        let mut data_ptr = 0;
+        value.ssbh_write(&mut writer, &mut data_ptr).unwrap();
+
+        // Null and empty arrays seem to use 0 offset and 0 length.
+        assert_eq!(
+            writer.into_inner(),
+            hex!("00000000 00000000 00000000 00000000")
+        );
+        assert_eq!(16, data_ptr);
+    }
+
+    #[test]
+    fn write_byte_buffer() {
+        let value = SsbhByteBuffer::new(vec![1u8, 2u8, 3u8, 4u8, 5u8]);
+
+        let mut writer = Cursor::new(Vec::new());
+        let mut data_ptr = 0;
+        value.ssbh_write(&mut writer, &mut data_ptr).unwrap();
+
+        assert_eq!(
+            writer.into_inner(),
+            hex!("10000000 00000000 05000000 00000000 01020304 05")
+        );
+        assert_eq!(21, data_ptr);
+    }
+
+    #[test]
+    fn write_vec() {
+        let value = vec![1u8, 2u8, 3u8, 4u8, 5u8];
+
+        let mut writer = Cursor::new(Vec::new());
+        let mut data_ptr = 0;
+        value.ssbh_write(&mut writer, &mut data_ptr).unwrap();
+
+        assert_eq!(writer.into_inner(), hex!("01020304 05"));
+        assert_eq!(5, data_ptr);
+    }
+
+    #[test]
+    fn write_empty_byte_buffer() {
+        let value = SsbhByteBuffer::new(Vec::new());
+
+        let mut writer = Cursor::new(Vec::new());
+        let mut data_ptr = 0;
+        value.ssbh_write(&mut writer, &mut data_ptr).unwrap();
+
+        // Null and empty arrays seem to use 0 offset and 0 length.
+        assert_eq!(
+            writer.into_inner(),
+            hex!("00000000 00000000 00000000 00000000")
+        );
+        assert_eq!(16, data_ptr);
+    }
 }
