@@ -1,19 +1,25 @@
 //! The [Matl] format stores a collection of materials used for model rendering.
 //! These files typically use the ".numatb" suffix like "model.numatb".
+//!
 //! The materials define some of the inputs for the specified shader and provide additional configuration over the rendering pipeline such as alpha blending settings.
 //! The materials in the [Matl] file are assigned to objects in the [Mesh](crate::formats::mesh::Mesh) file by the [Modl](crate::formats::modl::Modl) file.
 
 use crate::{Color4f, SsbhString, Vector4};
 use crate::{SsbhArray, SsbhEnum64};
 use binread::BinRead;
-#[cfg(feature = "serde")]
-use serde::{Deserialize, Serialize};
 use ssbh_write::SsbhWrite;
 
+#[cfg(feature = "serde")]
+use serde::{Deserialize, Serialize};
+
+#[cfg(feature = "strum")]
+use strum::{Display, EnumString, EnumVariantNames, FromRepr};
+
+// TODO: Rename these to Parameters to be consistent?
 /// A named material value.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-#[derive(BinRead, Debug, SsbhWrite)]
+#[derive(BinRead, Debug, SsbhWrite, PartialEq)]
 pub struct MatlAttributeV15 {
     /// Determines how the value in [param](#structfield.param) will be used by the shader.
     pub param_id: ParamId,
@@ -24,7 +30,7 @@ pub struct MatlAttributeV15 {
 /// A named material value.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-#[derive(BinRead, Debug, SsbhWrite)]
+#[derive(BinRead, Debug, SsbhWrite, PartialEq)]
 pub struct MatlAttributeV16 {
     /// Determines how the value in [param](#structfield.param) will be used by the shader.
     pub param_id: ParamId,
@@ -32,30 +38,17 @@ pub struct MatlAttributeV16 {
     pub param: SsbhEnum64<ParamV16>,
 }
 
-#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
-#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-#[derive(BinRead, Debug, SsbhWrite)]
-#[br(import(major_version: u16, minor_version: u16))]
-pub enum MatlAttributes {
-    #[br(pre_assert(major_version == 1 &&  minor_version == 5))]
-    Attributes15(SsbhArray<MatlAttributeV15>),
-    #[br(pre_assert(major_version == 1 &&  minor_version == 6))]
-    Attributes16(SsbhArray<MatlAttributeV16>),
-}
-
 /// A named collection of material values for a specified shader.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-#[derive(BinRead, Debug, SsbhWrite)]
-#[br(import(major_version: u16, minor_version: u16))]
-pub struct MatlEntry {
+#[derive(BinRead, Debug, SsbhWrite, PartialEq)]
+pub struct MatlEntryV15 {
     /// The name of this material.
     /// Material names should be unique.
     pub material_label: SsbhString,
 
     /// The collection of named material values.
-    #[br(args(major_version, minor_version))]
-    pub attributes: MatlAttributes,
+    pub attributes: SsbhArray<MatlAttributeV15>,
 
     /// The ID of the shader to associate with this material.
     /// For Smash Ultimate, the format is `<shader ID>_<render pass>`.
@@ -63,8 +56,37 @@ pub struct MatlEntry {
     pub shader_label: SsbhString,
 }
 
+/// A named collection of material values for a specified shader.
+#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(BinRead, Debug, SsbhWrite, PartialEq)]
+pub struct MatlEntryV16 {
+    /// The name of this material.
+    /// Material names should be unique.
+    pub material_label: SsbhString,
+
+    /// The collection of named material values.
+    pub attributes: SsbhArray<MatlAttributeV16>,
+
+    /// The ID of the shader to associate with this material.
+    /// For Smash Ultimate, the format is `<shader ID>_<render pass>`.
+    /// For example, the [shader_label](#structfield.shader_label) for shader `SFX_PBS_010002000800824f` and the `nu::opaque` render pass is "SFX_PBS_010002000800824f_opaque".
+    pub shader_label: SsbhString,
+}
+
+#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(BinRead, Debug, SsbhWrite)]
+#[br(import(major_version: u16, minor_version: u16))]
+pub enum MatlEntries {
+    #[br(pre_assert(major_version == 1 &&  minor_version == 5))]
+    EntriesV15(SsbhArray<MatlEntryV15>),
+    #[br(pre_assert(major_version == 1 &&  minor_version == 6))]
+    EntriesV16(SsbhArray<MatlEntryV16>),
+}
+
 /// A container of materials.
-/// Compatible with file version 1.6.
+/// Compatible with file version 1.5 and 1.6.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(BinRead, Debug, SsbhWrite)]
@@ -73,13 +95,13 @@ pub struct Matl {
     pub minor_version: u16,
     /// The material collection.
     #[br(args(major_version, minor_version))]
-    pub entries: SsbhArray<MatlEntry>,
+    pub entries: MatlEntries,
 }
 
 /// A material parameter value.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-#[derive(BinRead, Debug, SsbhWrite)]
+#[derive(BinRead, Debug, SsbhWrite, PartialEq)]
 #[br(import(data_type: u64))]
 pub enum ParamV15 {
     #[br(pre_assert(data_type == 1u64))]
@@ -117,7 +139,7 @@ pub enum ParamV15 {
 /// A material parameter value.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-#[derive(BinRead, Debug, SsbhWrite)]
+#[derive(BinRead, Debug, SsbhWrite, PartialEq)]
 #[br(import(data_type: u64))]
 pub enum ParamV16 {
     #[br(pre_assert(data_type == 1u64))]
@@ -152,11 +174,15 @@ pub enum ParamV16 {
     RasterizerState(MatlRasterizerStateV16),
 }
 
-/// The possible values for [param_id](struct.MatlAttribute.html.#structfield.param_id).
+/// An enumeration of all possible material parameters.
 /// Not all values are used by Smash Ultimate's shaders.
 /// For up to date documentation, see the [Material Parameters](https://github.com/ScanMountGoat/Smush-Material-Research/blob/master/Material%20Parameters.md) page on Github.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[cfg_attr(
+    feature = "strum",
+    derive(FromRepr, Display, EnumVariantNames, EnumString)
+)]
 #[derive(BinRead, Debug, SsbhWrite, Clone, Copy, PartialEq, Eq)]
 #[br(repr(u64))]
 #[ssbhwrite(repr(u64))]
@@ -531,8 +557,12 @@ pub enum ParamId {
     DiffuseLightingAoOffset = 365,
 }
 
-/// The possible values for [fill_mode](struct.MatlRasterizerStateV16.html.#structfield.fill_mode).
+/// Determines how polygons are shaded.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[cfg_attr(
+    feature = "strum",
+    derive(FromRepr, Display, EnumVariantNames, EnumString)
+)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(BinRead, Debug, SsbhWrite, Clone, Copy, PartialEq, Eq)]
 #[br(repr(u32))]
@@ -542,16 +572,20 @@ pub enum FillMode {
     Solid = 1,
 }
 
-/// The possible values for [cull_mode](struct.MatlRasterizerStateV16.html.#structfield.cull_mode).
+/// Determines the criteria for when to cull a face.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[cfg_attr(
+    feature = "strum",
+    derive(FromRepr, Display, EnumVariantNames, EnumString)
+)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-#[derive(BinRead, Debug, SsbhWrite, Clone, Copy, PartialEq)]
+#[derive(BinRead, Debug, SsbhWrite, Clone, Copy, PartialEq, Eq)]
 #[br(repr(u32))]
 #[ssbhwrite(repr(u32))]
 pub enum CullMode {
     Back = 0,
     Front = 1,
-    FrontAndBack = 2,
+    None = 2,
 }
 
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
@@ -567,9 +601,7 @@ pub struct MatlRasterizerStateV15 {
 #[derive(BinRead, Debug, Clone, PartialEq, SsbhWrite)]
 #[ssbhwrite(pad_after = 4)]
 pub struct MatlRasterizerStateV16 {
-    /// Determines the style for drawing polygon primitives.
     pub fill_mode: FillMode,
-    /// Determines the method of face culling to use.
     pub cull_mode: CullMode,
     pub depth_bias: f32,
     pub unk4: f32,
@@ -577,7 +609,13 @@ pub struct MatlRasterizerStateV16 {
     pub unk6: u32,
 }
 
+/// Determines how texture coordinates outside the 0 to 1 range
+/// are handled when sampling from the texture.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[cfg_attr(
+    feature = "strum",
+    derive(FromRepr, Display, EnumVariantNames, EnumString)
+)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(BinRead, Debug, SsbhWrite, Clone, Copy, PartialEq, Eq)]
 #[br(repr(u32))]
@@ -590,6 +628,10 @@ pub enum WrapMode {
 }
 
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[cfg_attr(
+    feature = "strum",
+    derive(FromRepr, Display, EnumVariantNames, EnumString)
+)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(BinRead, Debug, SsbhWrite, Clone, Copy, PartialEq, Eq)]
 #[br(repr(u32))]
@@ -601,6 +643,10 @@ pub enum MinFilter {
 }
 
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[cfg_attr(
+    feature = "strum",
+    derive(FromRepr, Display, EnumVariantNames, EnumString)
+)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(BinRead, Debug, SsbhWrite, Clone, Copy, PartialEq, Eq)]
 #[br(repr(u32))]
@@ -612,13 +658,17 @@ pub enum MagFilter {
 }
 
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[cfg_attr(
+    feature = "strum",
+    derive(FromRepr, Display, EnumVariantNames, EnumString)
+)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(BinRead, Debug, SsbhWrite, Clone, Copy, PartialEq, Eq)]
 #[br(repr(u32))]
 #[ssbhwrite(repr(u32))]
 pub enum FilteringType {
     Default = 0,
-    Default2 = 1,
+    Default2 = 1, // TODO: Does this change anything?
     AnisotropicFiltering = 2,
 }
 
@@ -636,14 +686,31 @@ pub struct MatlSampler {
     pub unk11: u32,
     pub unk12: u32,
     pub lod_bias: f32,
-    pub max_anisotropy: u32,
+    pub max_anisotropy: MaxAnisotropy,
+}
+
+#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[cfg_attr(
+    feature = "strum",
+    derive(FromRepr, Display, EnumVariantNames, EnumString)
+)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
+#[derive(BinRead, Debug, SsbhWrite, Clone, Copy, PartialEq, Eq)]
+#[br(repr(u32))]
+#[ssbhwrite(repr(u32))]
+pub enum MaxAnisotropy {
+    One = 1,
+    Two = 2,
+    Four = 4,
+    Eight = 8,
+    Sixteen = 16,
 }
 
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(BinRead, Debug, Clone, PartialEq, SsbhWrite)]
 pub struct MatlUvTransform {
-    pub x: f32, // TODO: translation/scale?
+    pub x: f32, // TODO: this is probably the same as the anim data type
     pub y: f32,
     pub z: f32,
     pub w: f32,
@@ -651,6 +718,10 @@ pub struct MatlUvTransform {
 }
 
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[cfg_attr(
+    feature = "strum",
+    derive(FromRepr, Display, EnumVariantNames, EnumString)
+)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[derive(BinRead, Debug, SsbhWrite, Clone, Copy, PartialEq, Eq)]
 #[br(repr(u32))]
@@ -672,7 +743,7 @@ pub enum BlendFactor {
 /// Determines the alpha blending settings to use when rendering.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-#[derive(BinRead, Debug, Clone, Copy, PartialEq, SsbhWrite)]
+#[derive(BinRead, Debug, Clone, PartialEq, SsbhWrite)]
 pub struct MatlBlendStateV15 {
     pub unk1: u64,
     pub unk2: u32,
@@ -688,7 +759,7 @@ pub struct MatlBlendStateV15 {
 /// Determines the alpha blending settings to use when rendering.
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-#[derive(BinRead, Debug, Clone, Copy, PartialEq, SsbhWrite)]
+#[derive(BinRead, Debug, Clone, PartialEq, SsbhWrite)]
 #[ssbhwrite(pad_after = 8)]
 pub struct MatlBlendStateV16 {
     pub source_color: BlendFactor,
@@ -697,7 +768,8 @@ pub struct MatlBlendStateV16 {
     pub unk4: u32,
     pub unk5: u32,
     pub unk6: u32,
-    pub unk7: u32,
+    /// 1 = enabled, 0 = disabled
+    pub alpha_sample_to_coverage: u32,
     pub unk8: u32,
     pub unk9: u32,
     pub unk10: u32,
